# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Tuple, List
from model_compression_toolkit.common.graph.base_graph import Graph
from model_compression_toolkit.common.graph.base_node import BaseNode


def get_compare_points(input_graph: Graph) -> Tuple[List[BaseNode], List[str], List, List]:
    """
    Create a list of nodes with weights in a graph and a corresponding list
    of their names for tensors comparison purposes. Also outputs 2 list of activations
    prior information collected from batch normalization nodes (if exists)
    Args:
        input_graph: Graph to get its points to compare.

    Returns:
        A list of nodes in a graph
        A list of the their names.
        A list of nodes mean collected from BatchNorms in the graph
        A list of nodes std collected from BatchNorms in the graph
    """
    compare_points = []
    compare_points_mean = []
    compare_points_std = []
    compare_points_name = []
    for n in input_graph.get_topo_sorted_nodes():
        if len(n.weights) > 0:
            compare_points.append(n)
            compare_points_name.append(n.name)
            compare_points_std.append(n.prior_info.std_output)
            compare_points_mean.append(n.prior_info.mean_output)
    return compare_points, compare_points_name, compare_points_mean, compare_points_std



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import numpy as np

from model_compression_toolkit.common.collectors.base_collector import BaseCollector
from model_compression_toolkit.common.framework_info import ChannelAxis


class MinMaxPerChannelCollector(BaseCollector):
    """
    Class to collect observed mix/max values of tensors that goes through it (passed to update).
    """

    def __init__(self,
                 axis: ChannelAxis,
                 init_min_value: float = None,
                 init_max_value: float = None):
        """
        Instantiate a collector for collecting min/max values of tensor per-channel.
        Args:
            axis: Compute the min/max values with regard to this axis.
            init_max_value: Initial maximal output value.
            init_min_value: Initial minimal output value.
        """
        super().__init__()
        self.axis = axis
        self.state = None
        self.init_min_value = init_min_value
        self.init_max_value = init_max_value
        self.ignore_init_values = False

    def scale(self, scale_factor: np.ndarray):
        """
        Scale all statistics in collector by some value.
        Since min/max are collected per-channel, they can be scaled either by a single factor or a
        scaling factor per-channel.

        Args:
            scale_factor: Factor to scale all collector's statistics by.

        """

        scale_per_channel = scale_factor.flatten().shape[0] > 1  # current scaling is per channel or not
        self.state = np.transpose(np.transpose(self.state) * scale_factor)

        if scale_per_channel:
            self.ignore_init_values = True
        else:
            if self.init_max_value is not None:
                self.init_max_value *= scale_factor
            if self.init_min_value is not None:
                self.init_min_value *= scale_factor

    def shift(self, shift_value: np.ndarray):
        """
        Shift all statistics in collector by some value.
        Since min/max are collected per-channel, they can be shifted either by a single value or a
        shifting value per-channel.

        Args:
            shift_value: Value to shift all collector's statistics by.
        """

        shift_per_channel = shift_value.flatten().shape[0] > 1  # current shifting is per channel or not
        self.state = np.transpose(np.transpose(self.state) + shift_value)

        if shift_per_channel:
            self.ignore_init_values = True
        else:
            if self.init_max_value is not None:
                self.init_max_value += shift_value
            if self.init_min_value is not None:
                self.init_min_value += shift_value

    @property
    def min(self) -> float:
        """
        Returns: Minimal value the collector observed (in general, not per-channel).
        """

        self.validate_data_correctness()
        return None if self.state is None else np.min(self.state[:, 1])

    @property
    def max(self) -> float:
        """
        Returns: Maximal value the collector observed (in general, not per-channel).
        """

        self.validate_data_correctness()
        return None if self.state is None else np.max(self.state[:, 0])

    @property
    def max_per_channel(self) -> np.ndarray:
        """
        Returns: Maximal value the collector observed per-channel.
        """

        self.validate_data_correctness()
        return self.state[:, 0]

    @property
    def min_per_channel(self) -> np.ndarray:
        """
        Returns: Minimal value the collector observed per-channel.
        """

        self.validate_data_correctness()
        return self.state[:, 1]

    def update(self,
               x: np.ndarray):
        """
        Update the min/max values the collector holds using a new tensor x to consider.

        Args:
            x: Tensor that goes through the collector and needs to be considered in the min/max computation.
        """

        axis = (len(x.shape) - 1) if self.axis == ChannelAxis.NHWC else self.axis.NCHW.value  # convert
        n = x.shape[axis]
        transpose_index = [axis, *[i for i in range(len(x.shape)) if i != axis]]
        x_reshape = np.reshape(np.transpose(x, transpose_index), [n, -1])
        if self.state is None:
            x_max = np.max(x_reshape, axis=-1)
            x_min = np.min(x_reshape, axis=-1)
        else:
            x_max = np.maximum(np.max(x_reshape, axis=-1), self.state[:, 0])
            x_min = np.minimum(np.min(x_reshape, axis=-1), self.state[:, 1])
        self.state = np.stack([x_max, x_min], axis=-1)



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numpy as np
from tensorflow.keras.layers import BatchNormalization, DepthwiseConv2D, Conv2DTranspose, Conv2D

from model_compression_toolkit.common import BaseNode
from model_compression_toolkit.common.graph.graph_matchers import NodeOperationMatcher, \
    NodeFrameworkAttrMatcher
from model_compression_toolkit.common.substitutions.batchnorm_folding import BatchNormalizationFolding
from model_compression_toolkit.keras.constants import KERNEL, LINEAR, ACTIVATION, DEPTHWISE_KERNEL, BIAS, GAMMA, BETA, \
    MOVING_MEAN, MOVING_VARIANCE, EPSILON, USE_BIAS, LAYER_NAME


def batchnorm_folding_node_matchers():
    """
    Function generates matchers for matching:
    (DepthwiseConv2D, Conv2D, Conv2DTranspose)[activation=linear] -> BatchNormalization.

    Returns:
        Matcher for batch norm nodes, and source nodes.
    """
    bn_node = NodeOperationMatcher(BatchNormalization)
    conv_node = NodeOperationMatcher(DepthwiseConv2D) | \
                  NodeOperationMatcher(Conv2D) | \
                  NodeOperationMatcher(Conv2DTranspose)

    activation_linear = NodeFrameworkAttrMatcher(ACTIVATION, LINEAR)
    source_node = conv_node & activation_linear
    return bn_node, source_node


def update_kernel_for_bn_folding_fn(conv_node: BaseNode,
                                    kernel: np.ndarray,
                                    weights_scale):
    """
    Args:
        conv_node: Convolution node to update the weight/kernel.
        kernel: The Convolution node's weight
        weights_scale: Weight scale factor in which to multiply the conv node's weight.

    Returns:
        The modified convolution node's weight/kernel/
    """
    if conv_node.type == DepthwiseConv2D:
        kernel = kernel * weights_scale.reshape(1, 1, kernel.shape[-2], kernel.shape[-1])
    elif conv_node.type == Conv2DTranspose:
        kernel = kernel * weights_scale.reshape(1, 1, -1, 1)
    else:
        kernel = kernel * weights_scale.reshape(1, 1, 1, -1)

    if conv_node.type == DepthwiseConv2D:
        kernel_name = DEPTHWISE_KERNEL
    else:
        kernel_name = KERNEL

    return kernel, kernel_name


def keras_batchnorm_folding() -> BatchNormalizationFolding:
    """

    Returns:
        A BatchNormalizationFolding initialized for Keras models.
    """
    bn_node, source_node = batchnorm_folding_node_matchers()
    return BatchNormalizationFolding(source_node,
                                     bn_node,
                                     update_kernel_for_bn_folding_fn,
                                     KERNEL,
                                     BIAS,
                                     GAMMA,
                                     BETA,
                                     MOVING_MEAN,
                                     MOVING_VARIANCE,
                                     EPSILON,
                                     USE_BIAS,
                                     LAYER_NAME)



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import tensorflow as tf
import tensorflow_model_optimization.quantization.keras.graph_transformations.model_transformer as mt
from model_compression_toolkit.keras.quantizer.mixed_precision.input_layer_quantize_transform import \
    InputLayerMixedPrecisionTransform

# As from Tensorflow 2.6, keras is a separate package and some classes should be imported differently.
if tf.__version__ < "2.6":
    from tensorflow.keras.layers import Input
    from tensorflow.python.keras.layers.core import TFOpLambda, SlicingOpLambda
    from tensorflow.python.keras.engine.base_layer import TensorFlowOpLayer
    from tensorflow.python.keras.layers import Layer
else:
    from keras import Input
    from keras.layers.core import TFOpLambda, SlicingOpLambda
    from keras.engine.base_layer import TensorFlowOpLayer, Layer

from model_compression_toolkit.common.model_builder_mode import ModelBuilderMode
from tensorflow_model_optimization.python.core.quantization.keras.quantize_wrapper import QuantizeWrapper
from typing import Tuple, Any, Dict, List
from tensorflow.python.util.object_identity import Reference as TFReference
from model_compression_toolkit.common.graph.functional_node import FunctionalNode
from model_compression_toolkit.common.logger import Logger
from model_compression_toolkit import common
from model_compression_toolkit.common.gptq.gptq_config import GradientPTQConfig
from model_compression_toolkit.common.framework_info import FrameworkInfo
from model_compression_toolkit.keras.default_framework_info import DEFAULT_KERAS_INFO
from model_compression_toolkit.keras.quantizer.mixed_precision.quantization_config_factory import \
    quantization_config_builder_mixed_precision
from model_compression_toolkit.keras.quantizer.gradient_ptq.config_factory import quantization_config_builder_gptq
from model_compression_toolkit.common import BaseNode, Graph
from model_compression_toolkit.common.graph.edge import EDGE_SINK_INDEX
from model_compression_toolkit.keras.back2framework.instance_builder import OperationHandler
from model_compression_toolkit.keras.reader.connectivity_handler import OutTensor

# In tf2.3 fake quant node is implemented as TensorFlowOpLayer, while in tf2.4 as TFOpLambda.
FQ_NODE_OP_V2_3 = 'FakeQuantWithMinMaxVars'
FQ_NODE_OP_V2_4 = 'quantization.fake_quant_with_min_max_vars'
BATCH_INPUT_SHAPE = 'batch_input_shape'


def get_node_name_from_layer(layer: Layer) -> str:
    """
    Get a node's name from the layer it was built from. For TensorFlowOpLayer
    we remove the prefix "tf_op_layer".

    Args:
        layer: Keras Layer to get its corresponding node's name.

    Returns:
        Name of the node that was built from the passed layer.
    """

    name = layer.name
    if isinstance(layer, TensorFlowOpLayer):  # remove TF op layer prefix
        name = '_'.join(name.split('_')[3:])
    return name


def is_layer_fake_quant(layer: Layer) -> bool:
    """
    Check whether a Keras layer is a fake quantization layer or not.
    Args:
        layer: Layer to check if it's a fake quantization layer or not.

    Returns:
        Whether a Keras layer is a fake quantization layer or not.
    """
    # in tf2.3 fake quant node is implemented as TensorFlowOpLayer, while in tf2.4 as TFOpLambda
    return (isinstance(layer, TensorFlowOpLayer) and layer.node_def.op == FQ_NODE_OP_V2_3) or (
            isinstance(layer, TFOpLambda) and layer.symbol == FQ_NODE_OP_V2_4)


def build_input_tensors_list(node: BaseNode,
                             graph: Graph,
                             node_to_output_tensors_dict: Dict[BaseNode, List[TFReference]]) -> List[List[TFReference]]:
    """
    Given a node, build a list of input tensors the node gets. The list is built
    based on the node's incoming edges and previous nodes' output tensors.

    Args:
        node: Node to build its input tensors list.
        graph: Graph the node is in.
        node_to_output_tensors_dict: A dictionary from a node to its output tensors.

    Returns:
        A list of the node's input tensors.
    """

    input_tensors = []
    # Go over a sorted list of the node's incoming edges, and for each source node get its output tensors.
    # Append them in a result list.
    for ie in graph.incoming_edges(node, sort_by_attr=EDGE_SINK_INDEX):
        _input_tensors = [node_to_output_tensors_dict[ie.source_node][ie.source_index]]
        input_tensors.append(_input_tensors)
    return input_tensors


def run_operation(n: BaseNode,
                  input_tensors: List[List[TFReference]],
                  op_func: Layer,
                  input_nodes_to_input_tensors: Dict[BaseNode, Any],
                  mode: ModelBuilderMode = ModelBuilderMode.QUANTIZED) -> List[TFReference]:
    """
    Applying the layer (op_func) to the input tensors (input_tensors).
    If quantized is set to True, and the layer's corresponding node (n) has quantization
    attributes, an additional fake-quantization node is built and appended to the layer.

    Args:
        n: The corresponding node of the layer it runs.
        input_tensors: List of references to Keras tensors that are the layer's inputs.
        op_func: Layer to apply to the input tensors.
        input_nodes_to_input_tensors: A dictionary from a node to its input tensors.
        mode: model quantization mode from ModelBuilderMode

    Returns:
        A list of references to Keras tensors. The layer's output tensors after applying the
        layer to the input tensors.
    """

    if len(input_tensors) == 0:  # Placeholder handling
        out_tensors_of_n_float = input_nodes_to_input_tensors[n]
        out_tensors_of_n = out_tensors_of_n_float
        if n.is_activation_quantization_enabled():
            if mode in [ModelBuilderMode.QUANTIZED, ModelBuilderMode.GPTQ] and n.final_activation_quantization_cfg:
                # Adding a fake quant node to Input when in GPTQ mode because quantize_model doesn't quantize the
                # input layer
                out_tensors_of_n = n.final_activation_quantization_cfg.quantize_node_output(out_tensors_of_n_float)
            elif mode in [ModelBuilderMode.MIXEDPRECISION]:
                if n.is_all_activation_candidates_equal():
                    # otherwise, we want to use the float tensor when building the model for MP search
                    out_tensors_of_n = n.candidates_quantization_cfg[
                        0].activation_quantization_cfg.quantize_node_output(out_tensors_of_n_float)

    else:
        input_tensors = [tensor for tensor_list in input_tensors for tensor in tensor_list]  # flat list of lists
        # Build a functional node using its args
        if isinstance(n, FunctionalNode):
            if n.inputs_as_list:  # If the first argument should be a list of tensors:
                out_tensors_of_n_float = op_func(input_tensors, *n.op_call_args, **n.op_call_kwargs)
            else:  # If the input tensors should not be a list but iterated:
                out_tensors_of_n_float = op_func(*input_tensors, *n.op_call_args, **n.op_call_kwargs)
        else:
            # If operator expects a single input tensor, it cannot be a list as it should
            # have a dtype field.
            if len(input_tensors) == 1:
                input_tensors = input_tensors[0]
            out_tensors_of_n_float = op_func(input_tensors)
        out_tensors_of_n = out_tensors_of_n_float

        # Add a fake quant node if the node has an activation threshold.
        if n.is_activation_quantization_enabled():
            if out_tensors_of_n_float.dtype != tf.float32:
                Logger.critical(
                    f"Trying to quantize node {n.name} activation of type {out_tensors_of_n_float.dtype} "
                    f"which is not supported, expected type float32")
            if mode in [ModelBuilderMode.QUANTIZED, ModelBuilderMode.GPTQ] and n.final_activation_quantization_cfg:
                out_tensors_of_n = n.final_activation_quantization_cfg.quantize_node_output(out_tensors_of_n_float)
            elif mode in [ModelBuilderMode.MIXEDPRECISION]:
                if n.is_all_activation_candidates_equal():
                    # otherwise, we want to use the float tensor when building the model for MP search
                    out_tensors_of_n = n.candidates_quantization_cfg[
                        0].activation_quantization_cfg.quantize_node_output(out_tensors_of_n_float)

    return out_tensors_of_n, out_tensors_of_n_float


def convert_node2name(in_node_to_output_tensors_dict):
    node_name_to_outtensors = dict()
    for node, tensors in in_node_to_output_tensors_dict.items():
        node_name_to_outtensors[node.name] = tensors
    return node_name_to_outtensors


def model_builder(graph: common.Graph,
                  mode: ModelBuilderMode = ModelBuilderMode.QUANTIZED,
                  append2output=None,
                  fw_info: FrameworkInfo = DEFAULT_KERAS_INFO, gptq_config: GradientPTQConfig = None) -> Tuple[
    tf.keras.models.Model, Any]:
    """
    Build a Keras model from a graph representing the model.
    The model is built by converting the graph nodes to Keras layers and applying them sequentially to get the model
    output tensors. The output tensors list and an input tensors list, then use to build the model.
    When the model is not built in float mode, the graph is being transformed by additional substitutions.

    Args:
        graph: Graph to build its corresponding Keras model.
        mode: Building mode. Read ModelBuilderMode description for more info.
        append2output: List of nodes or OutTensor objects. In float building mode,
        when the list contains nodes, all output tensors of all nodes are set as the model outputs.
        fw_info: Framework information (e.g., mapping from layers to their attributes to quantize).
        This is needed when using MIXEDPRECISION or GPTQ mode for passing the kernel attributes to
        the QuanteWrapper we use in both of these cases.
        gptq_config: GPTQ Configuration class..

    Returns:
        A tuple of the model, and an UserInformation object.
    """
    if gptq_config is None and mode == ModelBuilderMode.GPTQ:
        Logger.exception("Building a model in GPTQ require GPTQ configuration as input")
    node_to_output_tensors_dict = dict()
    node_to_output_tensors_dict_float = dict()
    model_output_tensors = []

    # Build an OperationHandler to handle conversions from graph nodes to Keras operators.
    oh = OperationHandler(graph)

    # Create a list of output nodes with their tensors' indices that are model's output. When building
    # in float mode, if a node has multiple out tensors, the node should appear in append2output multiple
    # times, thus creating OutTensor for each output tensor.
    if append2output is not None:
        output_list = [OutTensor(n, 0) for n in append2output]
    else:
        output_list = graph.get_outputs()

    # Hold a dictionary from an input node to its corresponding input tensor. It is needed for when
    # building the model. Initially input nodes with input tensors are added to the dictionary,
    # as they're not added later.
    input_nodes_to_input_tensors = {inode: Input(inode.framework_attr[BATCH_INPUT_SHAPE][1:], name=inode.name) for
                                    inode in graph.get_inputs()}

    # Build a list of the model's input tensors. Switching from a dictionary to a list
    # to keep the tensors input order, since inputs in Graph are ordered by their indices.
    inputs_list = []
    for input_node in graph.get_inputs():
        inputs_list.append(input_nodes_to_input_tensors.get(input_node))

    # Build a dictionary from node to its output tensors, by applying the layers sequentially.
    for n in oh.node_sort:
        op_func = oh.get_node_op_function(n)  # Get node operation function
        input_tensors = build_input_tensors_list(n,
                                                 graph,
                                                 node_to_output_tensors_dict)  # Fetch Node inputs
        out_tensors_of_n, out_tensors_of_n_float = run_operation(n,  # Run node operation and fetch outputs
                                                                 input_tensors,
                                                                 op_func,
                                                                 input_nodes_to_input_tensors,
                                                                 mode)

        if isinstance(out_tensors_of_n, list):
            node_to_output_tensors_dict.update({n: out_tensors_of_n})
            node_to_output_tensors_dict_float.update({n: out_tensors_of_n_float})
        else:
            node_to_output_tensors_dict.update({n: [out_tensors_of_n]})
            node_to_output_tensors_dict_float.update({n: [out_tensors_of_n_float]})

    # convert node_to_output_tensors_dict keys to nodes' names since oh.node_sort contains different objects than
    # original graph nodes.
    node_name_to_outtensors = convert_node2name(node_to_output_tensors_dict)
    node_name_to_outtensors_float = convert_node2name(node_to_output_tensors_dict_float)

    for ot in output_list:
        if len(node_name_to_outtensors[ot.node.name]) == 1 or append2output is None:
            if mode == ModelBuilderMode.GPTQ:
                model_output_tensors.append(node_name_to_outtensors_float[ot.node.name][ot.node_out_index])
            else:
                model_output_tensors.append(node_name_to_outtensors[ot.node.name][ot.node_out_index])
        else:  # When building float model - we collect all outputs from all nodes regardless the actual model's outputs
            if mode == ModelBuilderMode.GPTQ:
                # In case of GPTQ output the float data for the loss and not quantized.
                model_output_tensors.append(node_name_to_outtensors_float[ot.node.name])
            else:
                model_output_tensors.append(node_name_to_outtensors[ot.node.name])

    # Build the model.
    model = tf.keras.Model(inputs=inputs_list, outputs=model_output_tensors)

    # In GPTQ mode, wrap each layer in a QuantizeWrapper containing QuantizeConfig
    # that's built using the node quantization attributes.
    if mode == ModelBuilderMode.GPTQ:
        def _quantize(layer):
            nodes = graph.find_node_by_name(get_node_name_from_layer(layer))
            if len(nodes) == 1:
                node = nodes[0]
                return QuantizeWrapper(layer, quantization_config_builder_gptq(node, fw_info, gptq_config))
            elif is_layer_fake_quant(layer):
                return layer
            else:
                raise Exception(
                    f"Mismatch between keras model and graph can't find node named: {get_node_name_from_layer(layer)}")

        # clone each layer in the model and apply _quantize to the layer.
        model = tf.keras.models.clone_model(model, input_tensors=None, clone_function=_quantize)

    # In MIXEDPRECISION mode, wrap each layer that can be configured with bitwidth
    # in a QuantizeWrapper containing QuantizeConfig that holds a quantizer that
    # stores the quantized weights using all possible bitwidths.
    elif mode == ModelBuilderMode.MIXEDPRECISION:
        conf_nodes_names = graph.get_configurable_sorted_nodes_names()

        def _quantize_multiple_nbits(layer):
            nodes = graph.find_node_by_name(get_node_name_from_layer(layer))
            if len(nodes) == 1:
                node = nodes[0]
                # Wrap only if its weights should be quantized
                if node.name in conf_nodes_names:
                    if node.layer_class in [TFOpLambda, SlicingOpLambda]:
                        Logger.critical(f"Activation mixed-precision is not supported for layers of type "
                                        f"{node.layer_class}. Please modify the TargetPlatformModel object, "
                                        f"such that layers of type {node.layer_class} "
                                        f"won't have more than one quantization configuration option.")
                    return QuantizeWrapper(layer, quantization_config_builder_mixed_precision(node, fw_info))
                return layer

            elif is_layer_fake_quant(layer):
                return layer
            else:
                raise Exception(
                    f'Mismatch between keras model and graph cant find node named: {get_node_name_from_layer(layer)}')

        # clone each layer in the model and apply _quantize to the layer.
        model = tf.keras.models.clone_model(model, input_tensors=None, clone_function=_quantize_multiple_nbits)

        # We use a model transformer to wrap the input layer with QuantizeWrapper,
        # to allow layer configuration to different bitwidths.
        # A model transformer allows to modify a layer in an existing model, by applying the given list of
        # transformers on the model (in this case,
        # we only apply single transformer - InputLayerQuantizeTransform)
        model_inputs = graph.get_inputs()
        input_transformer = mt.ModelTransformer(model, [InputLayerMixedPrecisionTransform(inp, fw_info)
                                                        for inp in model_inputs])
        model = input_transformer.transform()[0]

    # Models that were built in float or quantized mode, should not be modified anymore.
    elif mode == ModelBuilderMode.FLOAT or mode == ModelBuilderMode.QUANTIZED:
        pass
    else:
        common.Logger.exception(f'Unknown model mode: {mode}')

    return model, graph.user_info



from model_compression_toolkit.common.collectors.statistics_collector import is_number


class NodePriorInfo:
    """
    Class to wrap all prior information we have on a node.
    """

    def __init__(self,
                 min_output: float = None,
                 max_output: float = None,
                 mean_output: float = None,
                 std_output: float = None):
        """
        Initialize a NodePriorInfo object.

        Args:
            min_output: Minimal output value of the node.
            max_output: Maximal output value of the node.
            mean_output: Mean output value of the node.
            std_output: Std output value of the node.
        """

        self.min_output = min_output
        self.max_output = max_output
        self.mean_output = mean_output
        self.std_output = std_output

    def is_output_bounded(self) -> bool:
        """

        Returns: Whether the node's output is bounded within a known range or not.

        """
        return is_number(self.min_output) and is_number(self.max_output)



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import torch
from torch.nn import AvgPool2d, MaxPool2d
from torch.nn.functional import avg_pool2d, max_pool2d, interpolate

from model_compression_toolkit.common.target_platform import TargetPlatformCapabilities
from model_compression_toolkit.common.target_platform.targetplatform2framework import OperationsSetToLayers, \
    LayerFilterParams
from model_compression_toolkit.common.target_platform.targetplatform2framework.attribute_filter import Eq
from model_compression_toolkit.tpc_models.tflite import get_tflite_tp_model
import operator


def get_pytorch_tflite_model():
    tflite_tp_model = get_tflite_tp_model()
    tflite_torch = TargetPlatformCapabilities(tflite_tp_model, name='tflite_torch')

    with tflite_torch:
        OperationsSetToLayers("PreserveQuantizationParams", [AvgPool2d,
                                                             avg_pool2d,
                                                             torch.cat,
                                                             torch.concat,
                                                             MaxPool2d,
                                                             max_pool2d,
                                                             torch.mul,
                                                             torch.multiply,
                                                             torch.reshape,
                                                             LayerFilterParams(interpolate, mode='bilinear'),
                                                             torch.nn.ZeroPad2d,
                                                             torch.gather,
                                                             torch.transpose,
                                                             torch.maximum,
                                                             torch.max,
                                                             torch.minimum,
                                                             torch.min,
                                                             torch.nn.functional.pad,
                                                             torch.select])

        OperationsSetToLayers("FullyConnected", [torch.nn.Linear,
                                                 torch.nn.functional.linear])

        OperationsSetToLayers("L2Normalization", [LayerFilterParams(torch.nn.functional.normalize,
                                                                    Eq('p',2) | Eq('p',None))])

        OperationsSetToLayers("LogSoftmax", [torch.nn.LogSoftmax])

        OperationsSetToLayers("Tanh", [torch.nn.Tanh,
                                       torch.nn.functional.tanh])

        OperationsSetToLayers("Softmax", [torch.nn.Softmax,
                                          torch.nn.functional.softmax])

        OperationsSetToLayers("Logistic", [torch.nn.Sigmoid,
                                           torch.nn.functional.sigmoid])

        OperationsSetToLayers("Conv2d", [torch.nn.Conv2d,
                                         torch.nn.functional.conv2d])

        OperationsSetToLayers("Relu", [torch.relu,
                                       torch.nn.ReLU,
                                       torch.nn.ReLU6,
                                       torch.nn.functional.relu,
                                       torch.nn.functional.relu6,
                                       LayerFilterParams(torch.nn.Hardtanh, min_val=0, max_val=6),
                                       LayerFilterParams(torch.nn.functional.hardtanh, min_val=0, max_val=6)])

        OperationsSetToLayers("Elu", [torch.nn.ELU,
                                      torch.nn.functional.elu])

        OperationsSetToLayers("BatchNorm", [torch.nn.BatchNorm2d,
                                            torch.nn.functional.batch_norm])

        OperationsSetToLayers("Squeeze", [torch.squeeze])

        OperationsSetToLayers("Add", [operator.add,
                                      torch.add])

    return tflite_torch





# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import List, Any, Tuple, Callable, Type
import numpy as np
import torch
from torch.nn import Module

from model_compression_toolkit import QuantizationConfig, FrameworkInfo, common, GradientPTQConfig, \
    MixedPrecisionQuantizationConfig
from model_compression_toolkit.common import Graph, BaseNode
from model_compression_toolkit.common.collectors.statistics_collector import BaseStatsCollector
from model_compression_toolkit.common.collectors.statistics_collector_generator import create_stats_collector_for_node
from model_compression_toolkit.common.framework_implementation import FrameworkImplementation
from model_compression_toolkit.common.gptq.gptq_training import GPTQTrainer
from model_compression_toolkit.common.model_builder_mode import ModelBuilderMode
from model_compression_toolkit.common.node_prior_info import NodePriorInfo
from model_compression_toolkit.common.user_info import UserInformation
from model_compression_toolkit.pytorch.back2framework.model_builder import model_builder
from model_compression_toolkit.pytorch.default_framework_info import DEFAULT_PYTORCH_INFO
from model_compression_toolkit.pytorch.graph_substitutions.substitutions.batchnorm_folding import \
    pytorch_batchnorm_folding
from model_compression_toolkit.pytorch.graph_substitutions.substitutions.relu_bound_to_power_of_2 import \
    ReLUBoundToPowerOfTwo
from model_compression_toolkit.pytorch.graph_substitutions.substitutions.mark_activation import MarkActivation
from model_compression_toolkit.pytorch.graph_substitutions.substitutions.reshape_with_static_shapes import \
    ReshapeWithStaticShapes
from model_compression_toolkit.pytorch.graph_substitutions.substitutions.scale_equalization import ScaleEqualization, \
    ScaleEqualizationWithPad
from model_compression_toolkit.pytorch.graph_substitutions.substitutions.shift_negative_activation import \
    pytorch_apply_shift_negative_correction
from model_compression_toolkit.pytorch.mixed_precision.sensitivity_evaluation import get_sensitivity_evaluation
from model_compression_toolkit.pytorch.pytorch_node_prior_info import create_node_prior_info
from model_compression_toolkit.pytorch.reader.reader import model_reader
import model_compression_toolkit.pytorch.constants as pytorch_constants
from model_compression_toolkit.pytorch.utils import to_torch_tensor, torch_tensor_to_numpy
from model_compression_toolkit.pytorch.graph_substitutions.substitutions.softmax_shift import \
    pytorch_softmax_shift


class PytorchImplementation(FrameworkImplementation):
    """
    An class with implemented methods to support optimizing Pytorch models.
    """

    def __init__(self):
        super().__init__()

    @property
    def constants(self):
        """
        Returns: Module of Pytorch constants.
        """
        return pytorch_constants

    def to_numpy(self,
                 tensor: torch.Tensor) -> np.ndarray:
        """
        Convert a Pytorch tensor to a Numpy array.
        Args:
            tensor: Pytorch tensor.
        Returns:
            Numpy array converted from the input tensor.
        """
        return torch_tensor_to_numpy(tensor)

    def to_tensor(self, tensor: Any) -> torch.Tensor:
        """
        Convert a Numpy array to a framework's tensor.
        Args:
            tensor: Numpy array.
        Returns:
            Framework's tensor converted from the input Numpy array.
        """
        return to_torch_tensor(tensor)

    def model_reader(self,
                     module: Module,
                     representative_data_gen: Callable) -> Graph:
        """
        Convert a framework's module into a graph.
        Args:
            module: Framework's module.
            representative_data_gen (Callable): Dataset used for calibration.
        Returns:
            Graph representing the input module.
        """
        return model_reader(module, representative_data_gen, self.to_numpy, self.to_tensor)

    def model_builder(self,
                      graph: Graph,
                      mode: ModelBuilderMode,
                      append2output: List[Any] = None,
                      fw_info: FrameworkInfo = DEFAULT_PYTORCH_INFO,
                      gptq_config: GradientPTQConfig = None) -> Tuple[Module, UserInformation]:
        """
        Build a Pytorch module from a graph.
        The mode determines how the module should be build. append2output is a list of Nodes
        to set as the module outputs.
        Args:
            graph: Graph to build the module from it.
            mode: Mode for how to build the module.
            append2output: List of Nodes to set as the module's outputs.
            fw_info: FrameworkInfo object with information about the specific framework's module
            gptq_config: GPTQ configuration class
        Returns:
            A tuple of the Pytorch module that was built and an UserInformation object.
        """
        return model_builder(graph,
                             mode,
                             append2output,
                             fw_info,
                             gptq_config)

    def run_model_inference(self,
                            model: Any,
                            input_list: List[Any]) -> Tuple[Any]:
        """
        Run the model logic on the given the inputs.
        Args:
            model: Pytorch model.
            input_list: List of inputs for the model.
        Returns:
            The Pytorch model's output in numpy format.
        """
        return torch_tensor_to_numpy(model(*to_torch_tensor(input_list)))

    def shift_negative_correction(self,
                                  graph: Graph,
                                  qc: QuantizationConfig,
                                  fw_info: FrameworkInfo) -> Graph:
        """
        Apply shift negative correction (SNC) on a graph.
        Args:
            graph: Graph to apply SNC on.
            qc: Quantization configuration.
            fw_info: FrameworkInfo object with information about the specific framework's module.
        Returns:
            Graph after SNC.
        """
        return pytorch_apply_shift_negative_correction(graph,
                                                       qc,
                                                       fw_info)

    def attach_sc_to_node(self,
                          node: BaseNode,
                          output_channel_index: int) -> BaseStatsCollector:
        """
        Return a statistics collector that should be attached to a node's output
        during statistics collection.
        Args:
            node: Node to return its collector.
            output_channel_index: Index of output channels of layers in the model's framework.
        Returns:
            Statistics collector for the node.
        """
        return create_stats_collector_for_node(node,
                                               output_channel_index=output_channel_index)

    def get_substitutions_marking(self) -> List[common.BaseSubstitution]:
        """
        Returns: A list of the framework substitutions used for marking
        points we fuse.
        """
        return [MarkActivation()]

    def get_substitutions_channel_equalization(self,
                                               quant_config: QuantizationConfig,
                                               fw_info: FrameworkInfo) -> List[common.BaseSubstitution]:
        """
        Return a list of the framework substitutions used for channel equalization.

        Args:
            quant_config: QuantizationConfig to determine which substitutions to return.
            fw_info: FrameworkInfo object with information about the specific framework's model.

        Returns:
            A list of the framework substitutions used after we collect statistics.
        """
        substitutions_list = []
        if quant_config.activation_channel_equalization:
            substitutions_list.extend([ScaleEqualization(quant_config, fw_info),
                                       ScaleEqualizationWithPad(quant_config, fw_info)])
        return substitutions_list

    def get_substitutions_prepare_graph(self) -> List[common.BaseSubstitution]:
        """

        Returns: A list of the framework substitutions used before we collect the prior information.

        """
        return [ReshapeWithStaticShapes()]

    def get_substitutions_pre_statistics_collection(self,
                                                    quant_config: QuantizationConfig
                                                    ) -> List[common.BaseSubstitution]:
        """
        Args:
            quant_config: QuantizationConfig to determine which substitutions to return.

        Returns: A list of the framework substitutions used before we build a quantized module.
        """
        substitutions_list = [pytorch_batchnorm_folding()]
        if quant_config.relu_bound_to_power_of_2:
            substitutions_list.append(ReLUBoundToPowerOfTwo())
        return substitutions_list

    def get_substitutions_post_statistics_collection(self,
                                                     quant_config: QuantizationConfig) -> List[common.BaseSubstitution]:
        """
        Return a list of the framework substitutions used after we collect statistics.
        Args:
            quant_config: QuantizationConfig to determine which substitutions to return.
        Returns:
            A list of the framework substitutions used after we collect statistics.
        """
        substitutions_list = []
        if quant_config.softmax_shift:
            substitutions_list.append(pytorch_softmax_shift())
        if quant_config.input_scaling:
            raise Exception('Input scaling is currently not supported for Pytorch.')
        return substitutions_list

    def get_substitutions_pre_build(self) -> List[common.BaseSubstitution]:
        """
        Returns: A list of the framework substitutions used before we build a quantized module.
        """
        return []

    def get_gptq_trainer_obj(self) -> Type[GPTQTrainer]:
        """
        Returns: GPTQTrainer object
        """
        raise Exception('This feature is currently not yet available for Pytorch models. Work in progress.')

    def get_sensitivity_evaluation_fn(self,
                                      graph: Graph,
                                      quant_config: MixedPrecisionQuantizationConfig,
                                      metrics_weights: np.ndarray,
                                      representative_data_gen: Callable,
                                      fw_info: FrameworkInfo) -> Callable:
        """
        Create and return a function to compute a sensitivity metric for a mixed-precision
        configuration (comparing to the float Pytorch module).
        Args:
            graph: Graph to build it's float and mixed-precision Pytorch models.
            quant_config: QuantizationConfig of how the module should be quantized.
            metrics_weights: Array of weights to weight the sensitivity among different layers.
            representative_data_gen: Dataset to use for retrieving images for the models inputs.
            fw_info: FrameworkInfo object with information about the specific framework's module.
        Returns:
            A function that computes the metric.
        """
        return get_sensitivity_evaluation(graph,
                                          quant_config,
                                          metrics_weights,
                                          representative_data_gen,
                                          fw_info)

    def get_node_prior_info(self,
                            node: BaseNode,
                            fw_info: FrameworkInfo,
                            graph: Graph) -> NodePriorInfo:
        """
        Get a NodePriorInfo object for a node that represents a Pytorch layer.
        Args:
            node: Node to get its prior info.
            fw_info: Framework specific information needed to create the prior info of the node.
            graph: Graph to check the next node type.
        Returns:
            NodePriorInfo with information about the node.
        """

        return create_node_prior_info(node=node,
                                      fw_info=fw_info,
                                      graph=graph)



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import torch
from torch.nn import Conv2d, Linear, BatchNorm2d, ConvTranspose2d, Hardtanh, ReLU, ReLU6
from torch.nn.functional import relu, relu6, hardtanh

from model_compression_toolkit.common.target_platform.targetplatform2framework import \
    TargetPlatformCapabilities, LayerFilterParams
from model_compression_toolkit.common.target_platform.targetplatform2framework import \
    OperationsSetToLayers
from model_compression_toolkit.tpc_models.qnnpack import get_qnnpack_model


def get_qnnpack_pytorch():
    qnnpack_tp_model = get_qnnpack_model()
    qnnpack_pytorch = TargetPlatformCapabilities(qnnpack_tp_model,
                                                 name='qnnpack_pytorch')

    with qnnpack_pytorch:
        OperationsSetToLayers("Conv", [Conv2d,
                                       torch.nn.functional.conv2d,
                                       ConvTranspose2d,
                                       torch.nn.functional.conv_transpose2d])

        OperationsSetToLayers("Linear", [Linear])

        OperationsSetToLayers("BatchNorm", [BatchNorm2d])

        OperationsSetToLayers("Relu", [torch.relu,
                                       ReLU,
                                       ReLU6,
                                       relu,
                                       relu6,
                                       LayerFilterParams(Hardtanh, min_val=0),
                                       LayerFilterParams(hardtanh, min_val=0)])

    return qnnpack_pytorch



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Callable, List

from model_compression_toolkit import common
from model_compression_toolkit.common import Logger
from model_compression_toolkit.common.constants import TENSORFLOW
from model_compression_toolkit.common.gptq.gptq_config import GradientPTQConfig
from model_compression_toolkit.common.mixed_precision.kpi import KPI
from model_compression_toolkit.common.framework_info import FrameworkInfo
from model_compression_toolkit.common.network_editors.actions import EditRule
from model_compression_toolkit.common.mixed_precision.mixed_precision_quantization_config import \
    MixedPrecisionQuantizationConfig, DEFAULT_MIXEDPRECISION_CONFIG
from model_compression_toolkit.common.post_training_quantization import post_training_quantization
from model_compression_toolkit.common.quantization.quantization_config import QuantizationConfig
from model_compression_toolkit.common.quantization.quantization_config import DEFAULTCONFIG

import importlib

from model_compression_toolkit.common.target_platform.targetplatform2framework import TargetPlatformCapabilities


if importlib.util.find_spec("tensorflow") is not None\
        and importlib.util.find_spec("tensorflow_model_optimization") is not None:
    import tensorflow as tf
    from model_compression_toolkit.keras.default_framework_info import DEFAULT_KERAS_INFO
    from model_compression_toolkit.keras.keras_implementation import KerasImplementation
    from model_compression_toolkit.keras.keras_model_validation import KerasModelValidation
    from tensorflow.keras.models import Model
    from model_compression_toolkit.keras.gradient_ptq.gptq_loss import multiple_tensors_mse_loss
    from keras.optimizer_v2.optimizer_v2 import OptimizerV2
    from model_compression_toolkit.keras.constants import DEFAULT_TP_MODEL

    from model_compression_toolkit import get_target_platform_capabilities
    DEFAULT_KERAS_TPC = get_target_platform_capabilities(TENSORFLOW, DEFAULT_TP_MODEL)

    def get_keras_gptq_config(n_iter: int,
                              optimizer: OptimizerV2 = tf.keras.optimizers.Adam(),
                              loss: Callable = multiple_tensors_mse_loss,
                              log_function: Callable = None,
                              train_bias: bool = True):
        """
        Create a GradientPTQConfig instance for Keras models.

        args:
            n_iter (int): Number of iterations to fine-tune.
            optimizer (OptimizerV2): Keras optimizer to use for fine-tuning.
            loss (Callable): loss to use during fine-tuning. should accept 4 lists of tensors. 1st list of quantized tensors, the 2nd list is the float tensors, the 3rd is a list of quantized weights and the 4th is a list of float weights.
            log_function (Callable): Function to log information about the gptq process.
            train_bias (bool): Whether to update the bias during the the fine-tuning or not.

        returns:
            a GradientPTQConfig object to use when fine-tuning the quantized model using gptq.

        examples:
            Create a GradientPTQConfig to run for 5 iteration:

            >>> gptq_conf = get_keras_gptq_config(n_iter=5)

            To disable the biases training, one may set train_bias to false (enabled by default):

            >>> gptq_conf = get_keras_gptq_config(n_iter=5, train_bias=false)

            Other Tensorflow optimizers can be passed:

            >>> gptq_conf = get_keras_gptq_config(n_iter=3, optimizer=tf.keras.optimizers.Nadam())

            The configuration can be passed to :func:`~model_compression_toolkit.keras_post_training_quantization` in order to quantize a keras model using gptq.

        """

        return GradientPTQConfig(n_iter,
                                 optimizer,
                                 loss=loss,
                                 log_function=log_function,
                                 train_bias=train_bias)

    def keras_post_training_quantization(in_model: Model,
                                         representative_data_gen: Callable,
                                         n_iter: int = 500,
                                         quant_config: QuantizationConfig = DEFAULTCONFIG,
                                         fw_info: FrameworkInfo = DEFAULT_KERAS_INFO,
                                         network_editor: List[EditRule] = [],
                                         gptq_config: GradientPTQConfig = None,
                                         analyze_similarity: bool = False,
                                         target_platform_capabilities: TargetPlatformCapabilities = DEFAULT_KERAS_TPC):
        """
        Quantize a pretrained Keras model using post-training quantization. By default, the model is quantized
        using a symmetric constraint quantization thresholds (power of two) as defined in the default TargetPlatformCapabilities.
        The model is first optimized using several transformations (e.g. BatchNormalization folding to
        preceding layers). Then, using a given dataset, statistics (e.g. min/max, histogram, etc.) are
        being collected for each layer's output (and input, depends on the quantization configuration).
        Thresholds are then being calculated using the collected statistics and the model is quantized
        (both coefficients and activations by default).
        If a gptq_config is passed, the quantized weights are optimized using gradient based post
        training quantization by comparing points between the float and quantized models, and minimizing the observed
        loss.

        Args:
            in_model (Model): Keras model to quantize.
            representative_data_gen (Callable): Dataset used for calibration.
            n_iter (int): Number of calibration iterations to run.
            quant_config (QuantizationConfig): QuantizationConfig containing parameters of how the model should be quantized. `Default configuration. <https://github.com/sony/model_optimization/blob/21e21c95ca25a31874a5be7af9dd2dd5da8f3a10/model_compression_toolkit/common/quantization/quantization_config.py#L154>`_
            fw_info (FrameworkInfo): Information needed for quantization about the specific framework (e.g., kernel channels indices, groups of layers by how they should be quantized, etc.). `Default Keras info <https://github.com/sony/model_optimization/blob/main/model_compression_toolkit/keras/default_framework_info.py>`_
            network_editor (List[EditRule]): List of EditRules. Each EditRule consists of a node filter and an action to change quantization settings of the filtered nodes.
            gptq_config (GradientPTQConfig): Configuration for using gptq (e.g. optimizer).
            analyze_similarity (bool): Whether to plot similarity figures within TensorBoard (when logger is enabled) or not.
            target_platform_capabilities (TargetPlatformCapabilities): TargetPlatformCapabilities to optimize the Keras model according to. `Default Keras TPC <https://github.com/sony/model_optimization/blob/main/model_compression_toolkit/tpc_models/keras_tp_models/keras_default.py>`_

        Returns:
            A quantized model and information the user may need to handle the quantized model.

        Examples:
            Import a Keras model:

            >>> from tensorflow.keras.applications.mobilenet import MobileNet
            >>> model = MobileNet()

            Create a random dataset generator:

            >>> import numpy as np
            >>> def repr_datagen(): return [np.random.random((1,224,224,3))]

            Import mct and pass the model with the representative dataset generator to get a quantized model:

            >>> import model_compression_toolkit as mct
            >>> quantized_model, quantization_info = mct.keras_post_training_quantization(model, repr_datagen, n_iter=1)

        """
        KerasModelValidation(model=in_model,
                             fw_info=fw_info).validate()
        return post_training_quantization(in_model,
                                          representative_data_gen,
                                          n_iter,
                                          quant_config,
                                          fw_info,
                                          KerasImplementation(),
                                          target_platform_capabilities,
                                          network_editor,
                                          gptq_config,
                                          analyze_similarity)


    def keras_post_training_quantization_mixed_precision(in_model: Model,
                                                         representative_data_gen: Callable,
                                                         target_kpi: KPI,
                                                         n_iter: int = 500,
                                                         quant_config: MixedPrecisionQuantizationConfig = DEFAULT_MIXEDPRECISION_CONFIG,
                                                         fw_info: FrameworkInfo = DEFAULT_KERAS_INFO,
                                                         network_editor: List[EditRule] = [],
                                                         gptq_config: GradientPTQConfig = None,
                                                         analyze_similarity: bool = False,
                                                         target_platform_capabilities: TargetPlatformCapabilities = DEFAULT_KERAS_TPC):
        """
         Quantize a pretrained Keras model using post-training quantization. By default, the model is quantized
         using a symmetric constraint quantization thresholds (power of two) as defined in the default
         TargetPlatformCapabilities.
         The model is first optimized using several transformations (e.g. BatchNormalization folding to
         preceding layers). Then, using a given dataset, statistics (e.g. min/max, histogram, etc.) are
         being collected for each layer's output (and input, depends on the quantization configuration).
         For each possible bit width (per operator, as defined in the TargetPlatformCapabilities) a
         threshold is then being calculated using the collected statistics.
         Then, using an ILP solver we find a mixed-precision configuration, and set a bit width
         for each quantizer (for both activations and weights quantizers, by default).
         In order to limit the maximal model's size, a target KPI need to be passed after weights_memory
         or activation_memory (or both) is set (in bytes).
         The model is then quantized (both coefficients and activations by default).
         If gptq_config is passed, the quantized weights are optimized using gradient based post
         training quantization by comparing points between the float and quantized models, and minimizing the
         observed loss.
         Notice that this feature is experimental.

         Args:
             in_model (Model): Keras model to quantize.
             representative_data_gen (Callable): Dataset used for calibration.
             target_kpi (KPI): KPI object to limit the search of the mixed-precision configuration as desired.
             n_iter (int): Number of calibration iterations to run.
             quant_config (MixedPrecisionQuantizationConfig): QuantizationConfig containing parameters of how the model should be quantized.
             fw_info (FrameworkInfo): Information needed for quantization about the specific framework (e.g., kernel channels indices, groups of layers by how they should be quantized, etc.). `Default Keras info <https://github.com/sony/model_optimization/blob/main/model_compression_toolkit/keras/default_framework_info.py>`_
             network_editor (List[EditRule]): List of EditRules. Each EditRule consists of a node filter and an action to change quantization settings of the filtered nodes.
             gptq_config (GradientPTQConfig): Configuration for using GPTQ (e.g. optimizer).
             analyze_similarity (bool): Whether to plot similarity figures within TensorBoard (when logger is enabled) or not.
             target_platform_capabilities (TargetPlatformCapabilities): TargetPlatformCapabilities to optimize the Keras model according to. `Default Keras TPC <https://github.com/sony/model_optimization/blob/main/model_compression_toolkit/tpc_models/keras_tp_models/keras_default.py>`_


         Returns:
             A quantized model and information the user may need to handle the quantized model.

         Examples:
             Import MCT:

             >>> import model_compression_toolkit as mct

             Import a Keras model:

             >>> from tensorflow.keras.applications.mobilenet_v2 import MobileNetV2
             >>> model = MobileNetV2()

             Create a random dataset generator:

             >>> import numpy as np
             >>> def repr_datagen(): return [np.random.random((1,224,224,3))]

             Create a mixed-precision configuration, to quantize a model with different bitwidths for different layers.
             The candidates bitwidth for quantization should be defined in the target platform model:

             >>> config = mct.MixedPrecisionQuantizationConfig()

             Create a KPI object to limit our returned model's size. Note that this value affects only coefficients
             that should be quantized (for example, the kernel of Conv2D in Keras will be affected by this value,
             while the bias will not):

             >>> kpi = mct.KPI(model.count_params() * 0.75)  # About 0.75 of the model size when quantized with 8 bits.

             Pass the model, the representative dataset generator, the configuration and the target KPI to get a
             quantized model:

             >>> quantized_model, quantization_info = mct.keras_post_training_quantization_mixed_precision(model,repr_datagen, target_kpi=kpi, n_iter=10, quant_config=config)

             For more configuration options, please take a look at our `API documentation <https://sony.github.io/model_optimization/api/api_docs/modules/mixed_precision_quantization_config.html>`_.

         """
        KerasModelValidation(model=in_model,
                             fw_info=fw_info).validate()

        if not isinstance(quant_config, MixedPrecisionQuantizationConfig):
            common.Logger.error("Given quantization config to mixed-precision facade is not of type "
                                "MixedPrecisionQuantizationConfig. Please use keras_post_training_quantization API,"
                                "or pass a valid mixed precision configuration.")

        common.Logger.info("Using experimental mixed-precision quantization. "
                           "If you encounter an issue please file a bug.")

        return post_training_quantization(in_model,
                                          representative_data_gen,
                                          n_iter,
                                          quant_config,
                                          fw_info,
                                          KerasImplementation(),
                                          target_platform_capabilities,
                                          network_editor,
                                          gptq_config,
                                          analyze_similarity,
                                          target_kpi)

else:
    # If tensorflow or tensorflow_model_optimization are not installed,
    # we raise an exception when trying to use these functions.
    def keras_post_training_quantization(*args, **kwargs):
        Logger.critical('Installing tensorflow and tensorflow_model_optimization is mandatory '
                        'when using keras_post_training_quantization. '
                        'Could not find Tensorflow package.')

    def keras_post_training_quantization_mixed_precision(*args, **kwargs):
        Logger.critical('Installing tensorflow and tensorflow_model_optimization is mandatory '
                        'when using keras_post_training_quantization_mixed_precision. '
                        'Could not find Tensorflow package.')

    def get_keras_gptq_config(*args, **kwargs):
        Logger.critical('Installing tensorflow and tensorflow_model_optimization is mandatory '
                        'when using keras_post_training_quantization_mixed_precision. '
                        'Could not find Tensorflow package.')




# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from model_compression_toolkit.common.quantization.quantization_params_generation.kl_selection import kl_selection_tensor, kl_selection_histogram
from model_compression_toolkit.common.quantization.quantization_params_generation.lp_selection import lp_selection_histogram, \
    lp_selection_tensor
from model_compression_toolkit.common.quantization.quantization_params_generation.kmeans_params import kmeans_tensor
from model_compression_toolkit.common.quantization.quantization_params_generation.lut_kmeans_params import lut_kmeans_tensor
from model_compression_toolkit.common.quantization.quantization_params_generation.mae_selection import mae_selection_tensor, mae_selection_histogram
from model_compression_toolkit.common.quantization.quantization_params_generation.mse_selection import mse_selection_tensor, mse_selection_histogram
from model_compression_toolkit.common.quantization.quantization_params_generation.no_clipping import no_clipping_selection_tensor, no_clipping_selection_min_max, no_clipping_selection_histogram
from model_compression_toolkit.common.quantization.quantization_params_generation.symmetric_selection import symmetric_no_clipping_selection_min_max
from model_compression_toolkit.common.quantization.quantization_params_generation.uniform_selection import uniform_no_clipping_selection_min_max
from model_compression_toolkit.common.quantization.quantization_params_generation.outlier_filter import z_score_filter
from model_compression_toolkit.common.quantization.quantization_params_generation.qparams_search import \
    qparams_selection_tensor_search, qparams_selection_histogram_search, quantize_tensor



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from typing import Callable

from model_compression_toolkit.common.target_platform import OpQuantizationConfig
from model_compression_toolkit.common.quantization.node_quantization_config import BaseNodeNodeQuantizationConfig, \
    NodeWeightsQuantizationConfig, NodeActivationQuantizationConfig
from model_compression_toolkit.common.quantization.quantization_config import QuantizationConfig


##########################################
# Every node holds a quantization configuration
# for its weights quantization, and a different quantization
# configuration for its activation quantization configuration.
##########################################

class CandidateNodeQuantizationConfig(BaseNodeNodeQuantizationConfig):
    """
    Class for representing candidate node configuration, which includes weights and activation configuration combined.
    """
    def __init__(self,
                 qc: QuantizationConfig,
                 op_cfg: OpQuantizationConfig,
                 activation_quantization_fn: Callable,
                 activation_quantization_params_fn: Callable,
                 weights_quantization_fn: Callable,
                 weights_quantization_params_fn: Callable,
                 weights_channels_axis: int
                 ):
        """

        Args:
            qc: QuantizationConfig to create the node's config from.
            op_cfg: OpQuantizationConfig of the node with quantizers types to use when creating node quantization configuration.
            activation_quantization_fn: Function to use when quantizing the node's activations.
            activation_quantization_params_fn: Function to use when computing the threshold for quantizing a node's activations.
            weights_quantization_fn: Function to use when quantizing the node's weights.
            weights_quantization_params_fn:  Function to use when computing the threshold for quantizing a node's weights.
            weights_channels_axis: Axis to quantize a node's kernel when quantizing per-channel.
        """

        self.activation_quantization_cfg = NodeActivationQuantizationConfig(qc,
                                                                            op_cfg,
                                                                            activation_quantization_fn,
                                                                            activation_quantization_params_fn)

        self.weights_quantization_cfg = NodeWeightsQuantizationConfig(qc,
                                                                      op_cfg,
                                                                      weights_quantization_fn,
                                                                      weights_quantization_params_fn,
                                                                      weights_channels_axis)



# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from model_compression_toolkit import target_platform as tpc
from model_compression_toolkit.common.target_platform.op_quantization_config import QuantizationMethod


def get_tflite_tp_model():
    # Create a quantization config. A quantization configuration defines how an operator
    # should be quantized on the modeled hardware. In TFLite
    # activations quantization is asymmetric, and weights quantization is symmetric:
    # https://www.tensorflow.org/lite/performance/quantization_spec#symmetric_vs_asymmetric
    eight_bits = tpc.OpQuantizationConfig(
        activation_quantization_method=QuantizationMethod.UNIFORM,
        weights_quantization_method=QuantizationMethod.SYMMETRIC,
        activation_n_bits=8,
        weights_n_bits=8,
        weights_per_channel_threshold=True,
        enable_weights_quantization=True,
        enable_activation_quantization=True,
        quantization_preserving=False,
        fixed_scale=None,
        fixed_zero_point=None,
        weights_multiplier_nbits=None
    )

    # Create a QuantizationConfigOptions, which defines a set
    # of possible configurations to consider when quantizing a set of operations (in mixed-precision, for example).
    # If the QuantizationConfigOptions contains only one configuration,
    # this configuration will be used for the operation quantization:
    default_configuration_options = tpc.QuantizationConfigOptions([eight_bits])

    # Create a TargetPlatformModel and set its default quantization config.
    # This default configuration will be used for all operations
    # unless specified otherwise (see OperatorsSet, for example):
    tflite_model = tpc.TargetPlatformModel(default_configuration_options, name='tflite')

    # To start defining the model's components (such as operator sets, and fusing patterns),
    # use 'with' the target platform model instance, and create them as below:
    with tflite_model:
        # In TFLite, the quantized operator specifications constraint operators quantization
        # differently. For more details:
        # https://www.tensorflow.org/lite/performance/quantization_spec#int8_quantized_operator_specifications
        tpc.OperatorsSet("PreserveQuantizationParams",
                         tpc.get_default_quantization_config_options().clone_and_edit(
                                  quantization_preserving=True))

        fc = tpc.OperatorsSet("FullyConnected",
                              tpc.get_default_quantization_config_options().clone_and_edit(
                                       weights_per_channel_threshold=False))

        tpc.OperatorsSet("L2Normalization",
                         tpc.get_default_quantization_config_options().clone_and_edit(
                                  fixed_zero_point=0, fixed_scale=1 / 128))
        tpc.OperatorsSet("LogSoftmax",
                         tpc.get_default_quantization_config_options().clone_and_edit(
                                  fixed_zero_point=127, fixed_scale=16 / 256))
        tpc.OperatorsSet("Tanh",
                         tpc.get_default_quantization_config_options().clone_and_edit(
                                  fixed_zero_point=0, fixed_scale=1 / 128))
        tpc.OperatorsSet("Softmax",
                         tpc.get_default_quantization_config_options().clone_and_edit(
                                  fixed_zero_point=-128, fixed_scale=1 / 256))
        tpc.OperatorsSet("Logistic",
                         tpc.get_default_quantization_config_options().clone_and_edit(
                                  fixed_zero_point=-128, fixed_scale=1 / 256))

        conv2d = tpc.OperatorsSet("Conv2d")
        depthwise_conv2d = tpc.OperatorsSet("DepthwiseConv2D")
        kernel = tpc.OperatorSetConcat(conv2d, fc, depthwise_conv2d)

        relu = tpc.OperatorsSet("Relu")
        elu = tpc.OperatorsSet("Elu")
        activations_to_fuse = tpc.OperatorSetConcat(relu, elu)

        batch_norm = tpc.OperatorsSet("BatchNorm")
        bias_add = tpc.OperatorsSet("BiasAdd")
        add = tpc.OperatorsSet("Add")
        squeeze = tpc.OperatorsSet("Squeeze",
                                   qc_options=tpc.get_default_quantization_config_options().clone_and_edit(quantization_preserving=True))

        # Source: https://github.com/tensorflow/tensorflow/blob/master/tensorflow/core/grappler/optimizers/remapper
        tpc.Fusing([kernel, bias_add])
        tpc.Fusing([kernel, bias_add, activations_to_fuse])
        tpc.Fusing([conv2d, batch_norm, activations_to_fuse])
        tpc.Fusing([conv2d, squeeze, activations_to_fuse])
        tpc.Fusing([batch_norm, activations_to_fuse])
        tpc.Fusing([batch_norm, add, activations_to_fuse])

    return tflite_model




# Copyright 2022 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from torch.nn import Module
from typing import Callable, List

from model_compression_toolkit.common.framework_info import FrameworkInfo
from model_compression_toolkit.common.graph.base_graph import Graph
from model_compression_toolkit.common.mixed_precision.mixed_precision_quantization_config import \
    MixedPrecisionQuantizationConfig
from model_compression_toolkit.common.mixed_precision.sensitivity_evaluation_manager import \
    SensitivityEvaluationManager, compute_mp_distance_measure
from model_compression_toolkit.pytorch.back2framework.model_builder import model_builder
from model_compression_toolkit.pytorch.mixed_precision.mixed_precision_wrapper import PytorchMixedPrecisionWrapper
from model_compression_toolkit.pytorch.utils import to_torch_tensor


def get_sensitivity_evaluation(graph: Graph,
                               quant_config: MixedPrecisionQuantizationConfig,
                               metrics_weights_fn: Callable,
                               representative_data_gen: Callable,
                               fw_info: FrameworkInfo) -> Callable:
    """
    Create a function to compute the sensitivity metric of an MP model (the sensitivity
    is computed based on the similarity of the interest points' outputs between the MP model
    and the float model).
    First, we build an MP model (a model where layers that can be configured in different bitwidths use
    a PytorchMixedPrecisionWrapper) and a baseline model (a float model).
    Then, and based on the outputs of these two models (for some batches from the representative_data_gen),
    we build a function to measure the sensitivity of a change in a bitwidth of a model's layer.
    Args:
        graph: Graph to get its sensitivity evaluation for changes in bitwidths for different nodes.
        quant_config: MixedPrecisionQuantizationConfig containing parameters of how the model should be quantized.
        metrics_weights_fn: Function to compute weights for a weighted average over the distances (per layer).
        representative_data_gen: Dataset used for getting batches for inference.
        fw_info: Framework information (e.g., mapping from layers to their attributes to quantize).
    Returns:
        Function to compute the sensitivity metric.
    """
    # Initiate a SensitivityEvaluationManager that contains all relevant objects for computing the distance metric.
    # The SensitivityEvaluationManager object initiates an MP model and a baseline model to be used for
    # comparison in the distance computation.
    # It generates and stores a set of image batches for evaluation.
    # It also runs and stores the baseline model's inference on the generated batches.
    # the model_builder method passed to the manager is the Pytorch model builder.
    sem = SensitivityEvaluationManager(graph, fw_info, quant_config, representative_data_gen, model_builder,
                                       move_tensors_func=lambda l: list(map(lambda t: t.detach().cpu(), l)))

    # Casting images tensors to torch.Tensor and putting them on same model as device
    sem.images_batches = list(map(lambda in_arr: to_torch_tensor(in_arr), sem.images_batches))

    # Initiating baseline_tensors_list since it is not initiated in SensitivityEvaluationManager init.
    sem.init_baseline_tensors_list()

    def _compute_metric(mp_model_configuration: List[int],
                        node_idx: List[int] = None) -> float:
        """
        Compute the sensitivity metric of the MP model for a given configuration (the sensitivity
        is computed based on the similarity of the interest points' outputs between the MP model
        and the float model).
        Args:
            mp_model_configuration: Bitwidth configuration to use to configure the MP model.
            node_idx: A list of nodes' indices to configure (instead of using the entire mp_model_configuration).
        Returns:
            The sensitivity metric of the MP model for a given configuration.
        """

        # Configure MP model with the given configuration.
        _configure_bitwidths_pytorch_model(sem.model_mp,
                                           sem.sorted_configurable_nodes_names,
                                           mp_model_configuration,
                                           node_idx)

        # Compute the distance matrix
        distance_matrix = sem.build_distance_metrix()

        # Configure MP model back to the same configuration as the baseline model
        baseline_mp_configuration = [0] * len(mp_model_configuration)
        _configure_bitwidths_pytorch_model(sem.model_mp,
                                           sem.sorted_configurable_nodes_names,
                                           baseline_mp_configuration,
                                           node_idx)

        return compute_mp_distance_measure(distance_matrix, metrics_weights_fn)

    return _compute_metric


def _configure_bitwidths_pytorch_model(model_mp: Module,
                                       sorted_configurable_nodes_names: List[str],
                                       mp_model_configuration: List[int],
                                       node_idx: List[int]):
    """
    Configure a dynamic Pytorch model (namely, model with layers that their weights
    bitwidth can be configured using PytorchMixedPrecisionWrapper) using an MP
    model configuration mp_model_configuration.
    Args:
        model_mp: Dynamic Pytorch model to configure. Note: model_mp is a PytorchModelBuilder object which composed of
        PytorchMixedPrecisionWrapper modules for each layer in the original module that can be configured
        for different bitwidths.
        sorted_configurable_nodes_names: List of configurable nodes names sorted topology.
        mp_model_configuration: Configuration of bitwidth indices to set to the model.
        node_idx: List of nodes' indices to configure (the rest layers are configured as the baseline model).
    """
    # Configure model
    # Note: the last configurable layer must be included in the interest points for evaluating the metric,
    # otherwise, it would not be considered throughout the mp optimization search (since it would not
    # affect the metric value)
    model_mp_layers_names = [l[0] for l in list(model_mp.named_children())]
    if node_idx is not None:  # configure specific layers in the mp model
        for node_idx_to_configure in node_idx:
            node_name = f'{sorted_configurable_nodes_names[node_idx_to_configure]}'
            if node_name in model_mp_layers_names:
                current_layer = model_mp.get_submodule(target=node_name)
                _set_layer_to_bitwidth(current_layer, mp_model_configuration[node_idx_to_configure])
            else:
                raise Exception("The last configurable node is not included in the list of interest points for"
                                "sensitivity evaluation metric for the mixed-precision search.")

    else:  # use the entire mp_model_configuration to configure the model
        for node_idx_to_configure, bitwidth_idx in enumerate(mp_model_configuration):
            node_name = f'{sorted_configurable_nodes_names[node_idx_to_configure]}'
            if node_name in model_mp_layers_names:
                current_layer = model_mp.get_submodule(target=node_name)
                _set_layer_to_bitwidth(current_layer, mp_model_configuration[node_idx_to_configure])
            else:
                raise Exception("The last configurable node is not included in the list of interest points for"
                                "sensitivity evaluation metric for the mixed-precision search.")


def _set_layer_to_bitwidth(wrapped_layer: Module,
                           bitwidth_idx: int):
    """
    Configure a layer (which is wrapped in a PytorchMixedPrecisionWrapper and holds a model's layer (nn.Module))
    to work with a different bitwidth.
    The bitwidth_idx is the index of the quantized-weights the quantizer in the PytorchMixedPrecisionWrapper holds.
    Args:
        wrapped_layer: Layer to change its bitwidth.
        bitwidth_idx: Index of the bitwidth the layer should work with.
    """
    assert isinstance(wrapped_layer, PytorchMixedPrecisionWrapper)
    # Configure the quantize_config to use a different bitwidth
    # (in practice, to use a different already quantized kernel).
    wrapped_layer.set_active_weights(bitwidth_idx)
    wrapped_layer.set_active_activation_quantizer(bitwidth_idx)



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from typing import Tuple, List
import numpy as np

from model_compression_toolkit.common.constants import MIN_THRESHOLD, EPS


def power_of_two_constraint(x: np.ndarray,
                            min_threshold: float = MIN_THRESHOLD) -> np.ndarray:
    """
    Compute the power-of-two threshold for quantizing a tensor x. The threshold
    is determined by the maximal value of the tensor (or min_threshold, the greater one, if a
    minimal value needed to be enforced for the threshold calculation).

    Args:
        x: Tensor to compute its threshold.
        min_threshold: Value to use for threshold computation if the maximal value of x
        is smaller than it.

    Returns:
        A constrained threshold to use when quantizing tensor x.
    """

    return np.power(2.0, np.ceil(np.log2(np.maximum(x, min_threshold))))


def calculate_delta(threshold: np.ndarray,
                    n_bits: int = 8,
                    signed: bool = False) -> np.ndarray:
    """
    Compute the step size of quantized values given the threshold, number of bits
    and whether its signed or unsigned.

    Args:
        threshold: Threshold to compute the step size according to.
        n_bits: Number of bits to compute the step size according to.
        signed: Whether quantization range is signed or not.

    Returns:
        Step size of quantized values according to a threshold, signedness and number of bits.
    """


    return threshold / (2 ** (n_bits - int(signed)))


def calculate_min_max_values(threshold: np.ndarray,
                             n_bits: int = 8,
                             signed: bool = False) -> Tuple[np.ndarray, np.ndarray]:
    """
    Compute the min/max values of a quantization range according to the threshold,
    number of bits and whether its signed or unsigned.

    Args:
        threshold: Threshold of quantization range to compute its min/max values.
        n_bits: Number of bits used in the quantization.
        signed: Whether the quantization range is signed or not.

    Returns:
        Min/max values of quantization range.
    """

    delta = calculate_delta(threshold,
                            n_bits=n_bits,
                            signed=signed)

    # If unsigned: min=0, otherwise its -threshold
    min_value = int(signed) * -threshold
    max_value = threshold - delta

    return min_value, max_value


def quantize_tensor(tensor_data: np.ndarray,
                    threshold: np.ndarray,
                    n_bits: int,
                    signed: bool) -> np.ndarray:
    """
    Quantize a tensor according to given: threshold, number of bits, and whether
    quantization range is sign or unsigned.

    Args:
        tensor_data: Tensor values to quantize.
        threshold: Threshold for quantization ranges.
        n_bits: Number of bits to quantize the tensor.
        signed: Whether the tensor contains negative values or not.

    Returns:
        Quantized data.
    """

    # Compute the step size of quantized values.
    delta = calculate_delta(threshold,
                            n_bits,
                            signed=signed)

    range_min = -threshold * int(signed)
    range_max = threshold - delta

    # Quantize the data between min/max of quantization range.
    return uniform_quantize_tensor(tensor_data,
                                   range_min=range_min,
                                   range_max=range_max,
                                   n_bits=n_bits)


def uniform_quantize_tensor(tensor_data: np.ndarray,
                            range_min: np.ndarray,
                            range_max: np.ndarray,
                            n_bits: int) -> np.ndarray:
    """
    Quantize a tensor according to given range (min, max) and number of bits.

    Args:
        tensor_data: Tensor values to quantize.
        range_min: minimum bound of the range for quantization (or array of min values per channel).
        range_max: maximum bound of the range for quantization (or array of max values per channel).
        n_bits: Number of bits to quantize the tensor.

    Returns:
        Quantized data.
    """

    # adjusts the quantization rage so the quantization grid include zero.
    a, b = fix_range_to_include_zero(range_min, range_max, n_bits)

    # Compute the step size of quantized values.
    delta = (b - a) / (2 ** n_bits - 1)

    # Clip data in range
    clipped_tensor = np.clip(tensor_data, a_min=a, a_max=b)

    # Quantize the data between min/max of quantization range.
    q = delta * np.round((clipped_tensor - a) / delta) + a
    return q


def kmeans_assign_clusters(cluster_centers: np.ndarray,
                           query: np.ndarray) -> np.ndarray:
    """
    Assign each data value in query with its closest cluster center point.
    Args:
        cluster_centers: the cluster centers to assign the query values.
        query: values for which to assign cluster centers.

    Returns: A tensor of indexes to the cluster centers that where assigned to each value in
             the query tensor.

    """
    d0 = query.shape[0]
    d1 = cluster_centers.shape[0]
    query_ = query.repeat(d1).reshape(d0, d1)
    cluster_centers_ = cluster_centers.repeat(d0).reshape(d1, d0).transpose(1, 0)
    return np.argmin(np.abs(query_ - cluster_centers_), axis=1)


def int_quantization_with_scale(data: np.ndarray,
                                scale: np.ndarray,
                                n_bits: int,
                                eps: float = EPS) -> np.ndarray:
    """
    Divides data by scale and quantizes it to integers in the range [2 ** (n_bits - 1) - 1, -2 ** (n_bits - 1)]
    Args:
        data: tensor data.
        scale: scale to divide the data.
        n_bits: number of bits that determines the quantization range.

    Returns:
        Quantized tensor.

    """
    return np.clip((data) / (scale + eps) * 2 ** (n_bits - 1),
                   a_max=2 ** (n_bits - 1) - 1, a_min=-2 ** (n_bits - 1))


def get_quantized_tensor(centers: np.ndarray,
                         scale: np.ndarray,
                         n_bits: int) -> np.ndarray:
    """
    Divides data by scale and quantizes it to integers in the range [2 ** (n_bits - 1) - 1, -2 ** (n_bits - 1)]
    Args:
        centers: centers points.
        scale: scale to divide the data.
        n_bits: number of bits that determines the quantization range.

    Returns:
        Quantized tensor.

    """
    return (np.round(centers) / (2 ** (n_bits - 1))) * scale


def get_tensor_max(tensor_data: np.ndarray,
                   per_channel: bool,
                   channel_axis: int) -> np.ndarray:
    """
    Returns the maximal value in the given tensor, or in each channel (if per_channel is True).
    Args:
        tensor_data: Tensor values to quantize.
        per_channel: Whether the quantization should be per-channel or not.
        channel_axis: Output channel index.

    Returns: maximal value (or values).

    """
    if per_channel:
        output_shape = get_output_shape(tensor_data.shape, channel_axis)
        tensor_data = reshape_tensor_for_per_channel_search(tensor_data, channel_axis)
        tensor_max = np.reshape(np.max(tensor_data, axis=-1), output_shape)
    else:
        tensor_max = np.max(tensor_data)
    return tensor_max


def get_tensor_min(tensor_data: np.ndarray,
                   per_channel: bool,
                   channel_axis: int) -> np.ndarray:
    """
    Returns the minimal value in the given tensor, or in each channel (if per_channel is True).
    Args:
        tensor_data: Tensor values to quantize.
        per_channel: Whether the quantization should be per-channel or not.
        channel_axis: Output channel index.

    Returns: minimal value (or values).

    """
    if per_channel:
        output_shape = get_output_shape(tensor_data.shape, channel_axis)
        tensor_data = reshape_tensor_for_per_channel_search(tensor_data, channel_axis)
        tensor_min = np.reshape(np.min(tensor_data, axis=-1), output_shape)
    else:
        tensor_min = np.min(tensor_data)
    return tensor_min


def reshape_tensor_for_per_channel_search(tensor_data: np.ndarray, channel_axis: int) -> np.ndarray:
    """
    Reshapes the given data to be compatible for search of quantization parameters for per-channel quantization.
    Args:
        tensor_data: Tensor values to quantize.
        channel_axis: Output channel index.

    Returns: A reshaped tensor.

    """
    # rearrange the shape indices for transposing the tensor
    shape_index = [channel_axis, *[i for i in range(len(tensor_data.shape)) if i is not channel_axis]]
    # New shape of the tensor after transposing it and reshape it
    new_shape = [tensor_data.shape[channel_axis], -1]
    tensor_data_t = np.transpose(tensor_data, shape_index)
    tensor_data_r = np.reshape(tensor_data_t, new_shape)
    return tensor_data_r


def fix_range_to_include_zero(range_min: np.ndarray,
                              range_max: np.ndarray,
                              n_bits: int) -> Tuple[np.ndarray, np.ndarray]:
    """
    Adjusting the quantization range to include representation of 0.0 in the quantization grid.
    If quantization per-channel, then range_min and range_max should be tensors in the specific shape that allows
    quantization along the channel_axis.

    Args:
        range_min: min bound of the quantization range (before adjustment).
        range_max: max bound of the quantization range (before adjustment).
        n_bits: Number of bits to quantize the tensor.

    Returns: adjusted quantization range
    """
    min_positive = range_min > 0
    max_negative = range_max < 0

    scale = (range_max - range_min) / (2 ** n_bits - 1)

    min_range_adj = scale * np.round(range_min / scale)
    max_range_adj = range_max - range_min + min_range_adj

    mid_range = np.logical_and(np.logical_not(min_positive), np.logical_not(max_negative))
    min_range_adj = min_range_adj * mid_range + max_negative * range_min
    max_range_adj = max_range_adj * mid_range + min_positive * range_max
    return min_range_adj, max_range_adj


def get_output_shape(tensor_shape, channel_axis):
    """
    Returns shape vector with the number of channels in the given channel_axis location and 1 at all other locations.
    Args:
        tensor_shape: A shape vector of a tensor.
        channel_axis: Output channel index.

    Returns: A shape vector of a tensor.

    """
    return [-1 if i is channel_axis else 1 for i in range(len(tensor_shape))]


def get_range_bounds(tensor_min, tensor_max):
    """
    Gets bounds on the quantization range limits for the minimization process.
    Calculates the bounds in a way that would leave a gap between the possible optimized values
    and the tensor min-max values.

    Args:
        tensor_min: min value of a tensor.
        tensor_max: max value of a tensor.

    Returns: An array with (lbound, ubound) pairs on the quantization range limit values.

    """
    # choosing bounds that have some gap from the original tensor min/max values.
    l_bound = tensor_min / 2 if tensor_min > 0 else tensor_min * 2
    u_bound = tensor_max * 2 if tensor_max > 0 else tensor_min / 2
    return [(l_bound, u_bound), (l_bound, u_bound)]


def get_threshold_bounds(min_threshold, max_threshold):
    """
    Gets bounds on the threshold for the minimization process.
    Calculates the bounds in a way that would leave a gap between the possible optimized threshold
    and the tensor max values. We use min_threshold as lower-bound to prevent the selected threshold
    from being zero or negative.

    Args:
        min_threshold: minimal threshold to use if threshold is too small (not used for this method).
        max_threshold: maximal threshold to be used in quantization.

    Returns: An array with a pair of (lbound, ubound) on the quantization threshold limit values.

    """
    max_threshold = max(min_threshold, max_threshold)
    return [(min_threshold, 2 * max_threshold)]



# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numpy as np
from typing import Tuple, Dict

from model_compression_toolkit.common.target_platform import QuantizationMethod
from model_compression_toolkit.common.collectors.statistics_collector import BaseStatsCollector
from model_compression_toolkit.common.constants import SIGNED
from model_compression_toolkit.common.quantization import quantization_params_generation
from model_compression_toolkit.common.node_prior_info import NodePriorInfo
from model_compression_toolkit.common.quantization.node_quantization_config import NodeActivationQuantizationConfig


def get_activations_qparams(activation_quant_cfg: NodeActivationQuantizationConfig,
                            nodes_prior_info: NodePriorInfo,
                            out_stats_container: BaseStatsCollector) -> Dict[str, float]:
    """
    Compute the activations params for a given node in a graph according to a params function.

    Args:
        activation_quant_cfg: node's activation quantization configuration.
        nodes_prior_info: Prior info collected for the node that is being quantized.
        out_stats_container: Tensor containing output statistics of the node.

    Returns:
        The computed activation quantization params.
    """

    bins_values, bins_counts = None, None

    # If the statistics container collected the histogram, we start by filtering outliers using z threshold
    # filtering, and then computing the threshold based on the filtered histogram.
    if out_stats_container.require_collection():
        bins_values, bins_counts = out_stats_container.hc.get_histogram()
        bins_counts = quantization_params_generation.z_score_filter(activation_quant_cfg.z_threshold,
                                                                    bins_values,
                                                                    bins_counts)
    min_value, max_value = out_stats_container.get_min_max_values()

    if nodes_prior_info.is_output_bounded():
        signed = min_value < 0
    else:
        signed = np.any(bins_values[:-1][bins_counts > 0] < 0)

    if nodes_prior_info.is_output_bounded():
        if activation_quant_cfg.activation_quantization_method == QuantizationMethod.POWER_OF_TWO:
            activation_quant_cfg.activation_quantization_params_fn = \
                quantization_params_generation.no_clipping_selection_min_max
        elif activation_quant_cfg.activation_quantization_method == QuantizationMethod.SYMMETRIC:
            activation_quant_cfg.activation_quantization_params_fn = \
                quantization_params_generation.symmetric_no_clipping_selection_min_max
        elif activation_quant_cfg.activation_quantization_method == QuantizationMethod.UNIFORM:
            activation_quant_cfg.activation_quantization_params_fn = \
                quantization_params_generation.uniform_no_clipping_selection_min_max

    activation_params = activation_quant_cfg.activation_quantization_params_fn(bins_values,
                                                                               bins_counts,
                                                                               activation_quant_cfg.l_p_value,
                                                                               activation_quant_cfg.activation_n_bits,
                                                                               min_value,
                                                                               max_value,
                                                                               min_threshold=activation_quant_cfg.min_threshold,
                                                                               quant_error_method=activation_quant_cfg.activation_error_method)
    activation_params.update({SIGNED: signed})

    return activation_params


# Copyright 2021 Sony Semiconductors Israel, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from enum import Enum

import numpy as np


class KPITarget(Enum):
    """
    Targets for which we define KPIs metrics for mixed-precision search.
    For each target that we care to consider in a mixed-precision search, there should be defined a set of
    kpi computation function, kpi aggregation function, and kpi target (within a KPI object).

    Whenever adding a kpi metric to KPI class we should add a matching target to this enum.

    WEIGHTS - KPI metric for weights quantization

    ACTIVATION - KPI metric for activation quantization

    """

    WEIGHTS = 'weights'
    ACTIVATION = 'activation'


class KPI(object):
    """
    Class to represent measurements of performance.
    """

    def __init__(self,
                 weights_memory: float = np.inf,
                 activation_memory: float = np.inf):
        """

        Args:
            weights_memory: Memory of a model's weights in bytes. Note that this includes only coefficients that should be quantized (for example, the kernel of Conv2D in Keras will be affected by this value, while the bias will not).
            activation_memory: Memory of a model's activation in bytes, according to the given activation kpi metric.
        """
        self.weights_memory = weights_memory
        self.activation_memory = activation_memory

    def __repr__(self):
        return f"Weights_memory: {self.weights_memory}, " \
               f"Activation_memory: {self.activation_memory}"
